/**
 * Components - MercuryAccordion
 * - 01 - MercuryAccordion
 */

/*------------------------------------*\
  01 - MercuryAccordion
\*------------------------------------*/

/**
 * MercuryAccordion class.
 *
 * Manages behavior of a single accordion.
 */
class MercuryAccordion {

  /**
   * Create a new MercuryAccordion instance.
   *
   * @param {HTMLElement} wrapperElement
   *  The wrapper element for the accordion.
   */
  constructor(wrapperElement) {
    this.wrapperElement = wrapperElement;
    this.triggerElement = wrapperElement.querySelector('.me-accordion__trigger');
    this.contentElement = wrapperElement.querySelector('.me-accordion__content');

    this.id = this.triggerElement.id;

    if (!this.id) {
      console.warn('MercuryAccordion was not initialized because the id attribute is missing from...', this.triggerElement);
      return;
    }

    // Make the MercuryAccordion instance available on the elements.
    this.wrapperElement.MercuryAccordion = this;
    this.triggerElement.MercuryAccordion = this;
    this.contentElement.MercuryAccordion = this;

    // Initialize the accordion.
    this.init();
  }

  /**
   * The CSS class name applied when the accordion is open.
   *
   * @type {string}
   * @static
   *
   * This class name will be applied to the wrapper element and the content element
   * when the accordion is open. This allows more styling flexibility than strictly
   * relying on the `aria-expanded` attribute of the trigger and is more reliable when
   * when dealling with nested accordions.
   */
  static OPEN_CLASS = 'is-open';

  /**
   * Store a set of all accordion ids.
   *
   * @type {Set<string>}
   * @static
   */
  static accordions = new Set();

  /**
   * Update the UI for all accordions.
   *
   * This method ensures the correct accordion is expanded
   * based on the session Storage state. Calling this method
   * will ensure accordions don't appear to close when content
   * is re-rendered.
   *
   * @static
   * @example
   * MercuryAccordion.updateUi();
   */
  static updateUi() {
    // Loop over all accordions and open or close them based on their state.
    MercuryAccordion.accordions.forEach((id) => {
      const mercuryAccordion = document.getElementById(id).MercuryAccordion;
      mercuryAccordion[mercuryAccordion.isOpen ? 'open' : 'close']();
    });
  }

  /**
   * Setter for the isOpen property.
   * This setter will store the state of the accordion in session storage.
   *
   * @param {boolean} isOpen - The state of the accordion.
   */
  set isOpen(isOpen) {
    sessionStorage.setItem(this.getKey(), isOpen);
  }

  /**
   * Getter for the isOpen property.
   * This getter will return the state of the accordion from session storage.
   */
  get isOpen() {
    return sessionStorage.getItem(this.getKey()) === 'true';
  }

  /**
   * Get the unique key for this instance
   *
   * @returns {string} - The key used to store the accordion state in session storage.
   */
  getKey() {
    return `MercuryAccordion.${this.id}`;
  }

  /**
   * Initialize the accordion.
   *
   * This function initializes the accordion by adding event listeners and
   * opens the accordion if the hash in the location matches the accordion id
   * or content within it.
   */
  init() {
    // Bind event listeners.
    this.triggerElement.addEventListener('click', this.onTriggerClick);
    this.triggerElement.addEventListener('keydown', this.onTriggerKeydown);
    window.addEventListener('hashchange', this.openFromHash);

    // Add the accordion to the list of accordions.
    MercuryAccordion.accordions.add(this.id);

    // Store the initial state of the accordion if not already set.
    if (sessionStorage.getItem(this.id) === null) {
      sessionStorage.setItem(this.id, this.triggerElement.getAttribute('aria-expanded') === 'true');
    }

    // Open or close the accordion based on the stored state.
    this.isOpen ? this.open() : this.close();

    this.openFromHash();
  }

  /**
   * Cleanup current MercuryAccordion instance.
   *
   * This function removes event listeners from the trigger element
   * and removes the accordion from the list of accordions.
   *
   * We do not want to remove the open state of the accordion from session storage
   * as it would defeat the purpose of storing the state in the first place.
   */
  destroy() {
    this.triggerElement.removeEventListener('click', this.onTriggerClick);
    this.triggerElement.removeEventListener('keydown', this.onTriggerKeydown);
    window.removeEventListener('hashchange', this.openFromHash);
    MercuryAccordion.accordions.delete(this.id);
  }

  /**
   * Collapses an accordion section.
   *
   * This function removes the `openedClass` from both the
   * accordion content and the accordion trigger elements.
   * Additionally, it sets the `aria-expanded` attribute of the
   * accordion trigger to 'false', indicating that the section is collapsed.
   *
   * @see onTriggerClick
   * @see onTriggerKeydown
   */
  close = () => {
    this.contentElement.classList.remove(MercuryAccordion.OPEN_CLASS);
    this.wrapperElement.classList.remove(MercuryAccordion.OPEN_CLASS);
    this.triggerElement.setAttribute('aria-expanded', 'false');
    this.isOpen = false;
  };

  /**
   * Expands the accordion section by adding the expanded CSS class
   * to the content and trigger elements, and sets the aria-expanded
   * attribute to true.
   *
   * This function modifies the DOM to visually and semantically
   * indicate that the accordion content is currently expanded.
   *
   * @see onTriggerClick
   * @see onTriggerKeydown
   */
  open() {
    this.contentElement.classList.add(MercuryAccordion.OPEN_CLASS);
    this.wrapperElement.classList.add(MercuryAccordion.OPEN_CLASS);
    this.triggerElement.setAttribute('aria-expanded', 'true');

    // Save the hash of the accordion to the current url.
    // This is done to facilitate deep linking to specific accordion sections.
    const hash = `#${this.id}`;
    if (history.pushState) {
      history.pushState(null, null, hash);
    }
    else {
      location.hash = hash;
    }

    this.isOpen = true;
  }

  /**
   * Open the accordion based on the hash in the location.
   *
   * This method ensures that the accordion is open if it or content
   * within it contains the ID in the location url hash.
   */
  openFromHash = () => {
    if (window.location.hash) {
      const hash = window.location.hash.substring(1);
      let element = document.getElementById(hash);
      if (element && (element === this.wrapperElement || this.wrapperElement.contains(element)) ) {
        this.open();
      }
    }
  }

  /**
   * Toggles the state of the accordion element open or closed.
   *
   * @see open
   * @see close
   * @see onTriggerClick
   * @see onTriggerKeydown
   */
  toggle() {
    this.isOpen ? this.close() : this.open();
  }

  /**
   * Handle the click event on the trigger.
   */
  onTriggerClick = (event) => {
    event.preventDefault();
    event.stopPropagation();
    this.toggle();
  }

  /**
   * Event handler for `keydown` events on a specific trigger element.
   *
   * This function is designed to handle keyboard interactions for improved accessibility,
   * specifically listening for the "Enter" and "Space" keys. When either of these keys
   * is pressed, it prevents the default action and toggles a component's state between
   * open and closed.
   *
   * @param {KeyboardEvent} event - The keyboard event object.
   *
   */
  onTriggerKeydown = (event) => {
    switch (event.key) {
      case 'Enter':
      case ' ':
        event.preventDefault();
        this.toggle();
        break;
    }
  }
}
