/**
 * Components - MercuryTabs
 * - 01 - MercuryTab
 * - 02 - MercuryTabs
 */


/*------------------------------------*\
  01 - MercuryTab
\*------------------------------------*/

/**
 * MercuryTab class.
 *
 * Manages behavior of a single tab within a tab group.
 */
class MercuryTab {

  /**
   * Create a new MercuryTab instance.
   *
   * @param {HTMLElement} triggerElement
   *  The trigger element for the tab.
   */
  constructor(triggerElement) {
    this.triggerElement = triggerElement;
    this.groupId = triggerElement.getAttribute('data-group-id');
    this.panelId = triggerElement.getAttribute('aria-controls');
    this.panelElement = document.getElementById(this.panelId);
    this.mercuryTabs = MercuryTabs.groups.get(this.groupId);
    this.panelElement.mercuryTabs = this.mercuryTabs;
    this.init();
  }

  /**
   * Initialize the tab.
   */
  init() {
    this.triggerElement.addEventListener('click', this.onTriggerClick);
    this.triggerElement.addEventListener('keydown', this.onTriggerKeydown);
    this.triggerElement.tabIndex = -1;
  }

  /**
   * Convenience method to focus on the trigger element.
   */
  focus() {
    this.triggerElement.focus();
  }

  /**
   * Is this tab the first in the group?
   *
   * @returns {boolean}
   *  True if this tab is the first in the group.
   */
  isFirst() {
    return this === this.triggers()[0];
  }

  /**
   * Is this tab the last in the group?
   *
   * @returns {boolean}
   *  True if this tab is the last in the group.
   */
  isLast() {
    return this === this.triggers().at(-1);
  }

  /**
   * Get the first trigger in the group.
   *
   * @returns {MercuryTab}
   *  The first trigger in the group.
   */
  firstTrigger() {
    return this.triggers()[0];
  }

  /**
   * Get the last trigger in the group.
   *
   * @returns {MercuryTab}
   *  The last trigger in the group.
   */
  lastTrigger() {
    return this.triggers().at(-1);
  }

  /**
   * Get the next trigger in the group.
   *
   * @returns {MercuryTab}
   *  The next trigger in the group.
   */
  nextTrigger() {
    const triggers = this.triggers();
    return triggers.at(triggers.indexOf(this) + 1);
  }

  /**
   * Get the previous trigger in the group.
   */
  prevTrigger() {
    const triggers = this.triggers();
    return triggers.at(triggers.indexOf(this) - 1);
  }

  /**
   * Set focus on an element on the next event loop to ensure it is visible.
   *
   * @param {HTMLElement} element
   *  The element to focus on.
   */
  setFocus( element ) {
    setTimeout(() => {
      element.focus();
    }, 0);
  }

  /**
   * Get the triggers for this tab group.
   *
   * @returns {Array<MercuryTab>}
   *  The triggers for this tab group.
   */
  triggers = () => {
    const group = MercuryTabs.groups.get(this.groupId);
    return group.triggers ? group.triggers : [];
  }

  /**
   * Handle the click event on the trigger.
   */
  onTriggerClick = (event) => {
    event.preventDefault();
    event.stopPropagation();
    this.open();
  }

  /**
   * Handle the keydown event on the trigger.
   */
  onTriggerKeydown = (event) => {
    switch (event.key) {
      case 'ArrowLeft':
        if (this.isFirst()) {
          this.setFocus(this.lastTrigger());
        } else {
          this.setFocus(this.prevTrigger());
        }
        break;
      case 'ArrowRight':
        if (this.isLast()) {
          this.setFocus(this.firstTrigger());
        } else {
          this.setFocus(this.nextTrigger());
        }
        break;
      case 'Home':
        this.setFocus(this.firstTrigger());
        break;
      case 'End':
        this.setFocus(this.lastTrigger());
        break;
      case 'Enter':
      case ' ':
        event.preventDefault();
        this.open();
        break;
    }
  }

  open() {
    this.mercuryTabs.open(this.panelId);
  }
}

/*------------------------------------*\
  02 - MercuryTabs
\*------------------------------------*/

/**
 * MercuryTabs class.
 *
 * This class is responsible for managing the state of MercuryTabs.
 *
 */
class MercuryTabs {

  /**
   * Create a new MercuryTabs instance.
   *
   * @param {HTMLElement} tabsElement
   *  The root element for the tabs.
   *
   */
  constructor(tabsElement) {

    // Ensure the element has a group id.
    this.groupId = tabsElement.getAttribute('data-group-id');
    if (!this.groupId) {
      console.warn('MercuryTabs was not initialized because the data-group-id attribute is missing from...', tabsElement);
      return;
    }
    this.tabsElement = tabsElement;
    this.init();

    // Attach the MercuryTabs instance to the HTMLElement.
    tabsElement.MercuryTabs = this;
  }

  /**
   * Store for all tab groups.
   *
   * @type {Map<string, MercuryTabs>}
   */
  static groups = new Map();

  /**
   * Open a tab based on the hash in the location.
   *
   * This method can be called on page load to ensure content
   * within a tab is deeplinked to any parent tabs are open.
   *
   * @static
   *
   * @example
   *   MercuryTabs.openTabFromHash();
   *
   */
  static openTabFromHash() {
    if (window.location.hash) {
      const hash = window.location.hash.substring(1);
      let element = document.getElementById(hash);
      if (element) {
        // Loop through all parent tabs and open them.
        while (element = element.closest('.me-tabs__tab')) {
          element.mercuryTabs.open(element.id, false);
          element = element.parentNode;
        }
      }
    }
  }


  /**
   * Setter for the activeTab property.
   *
   * This setter will store the activeTab in session storage.
   *
   * @param {string} id - The HTML ID of the active tab.
   */
  set activeTab(id) {
    sessionStorage.setItem(this.getKey(), id);
  }

  /**
   * Getter for the activeTab property.
   *
   * This getter will return the activeTab from session storage.
   */
  get activeTab() {
    return sessionStorage.getItem(this.getKey());
  }

  /**
   * Get the unique key for this instance.
   *
   * This key is used to store the active tab in session storage.
   */
  getKey() {
    return `MercuryTabs.${this.groupId}`;
  }

  /**
   * Cleanup current MercuryTabs instance.
   */
  destroy() {
    MercuryTabs.groups.delete(this.groupId);
  }

  /**
   * Handle tab button click event.
   *
   * @param {MouseEvent} event
   *  The click event.
   */
  onTriggerClick = (event) => {
    const itemId = event.target.getAttribute('aria-controls');
    this.open(itemId);
  }

  /**
   * Initialize the tabs.
   */
  init() {
    MercuryTabs.groups.set(this.groupId, this);

    this.triggers = [...this.tabsElement.querySelectorAll(`[role="tab"][data-group-id="${this.groupId}"]`)]
      .map(trigger => new MercuryTab(trigger));

    // Set a default active tab if one is not set.
    if (this.activeTab === null || document.getElementById(this.activeTab) === null) {
      if (this.triggers.length > 0) {
        this.activeTab = this.triggers[0].panelId;
      }
    }

    this.updateUi();
  }

  /**
   * Open a tab.
   *
   * @param {string} itemId
   *  The id of the tab to open.
   * @param {boolean} shouldUpdateHash true
   *  Whether or not to update the location hash with the itemId.
   *  Default is true.
   */
  open(itemId, shouldUpdateHash = true) {
    this.activeTab = itemId;
    this.updateUi();
    if (shouldUpdateHash) {
      this.updateHash(itemId);
    }
  }

  /**
   * Update the location hash with the itemId.
   * This allows users to deeplink to a specific tab.
   *
   * @param {string} id
   *  The id of the tab to update the hash with.
   */
  updateHash(id) {
    const hash = `#${id}`;
    if (history.pushState) {
      history.pushState(null, null, hash);
    }
    else {
      location.hash = hash;
    }
  }

  /**
   * Update the UI based on the current state.
   * This method ensures the correct tab is expanded
   * based on the current state.
   */
  updateUi () {
    this.tabsElement.querySelectorAll(`.me-tabs__tab-button[data-group-id="${this.groupId}"]`).forEach(button => {
      const itemId = button.getAttribute('aria-controls');
      const target = document.getElementById(itemId);
      const isExpanded = this.activeTab === itemId;

      if (isExpanded) {
        button.removeAttribute('tabIndex');
      } else {
        button.tabIndex = -1;
      }

      button.setAttribute('aria-selected', String(isExpanded));
      target.setAttribute('aria-hidden', String(!isExpanded));
    });
  }
}
