((Drupal, debounce, Sortable, once) => {
  /**
   * Save a label in the layout.
   */
  function saveLabel() {
    const container = document.querySelector('[data-editable-label]');
    if (!container) return; // Exit if the container is not found

    const uuidElement = container.closest('[data-uuid]');
    const layoutIdElement = container.closest('[data-lpb-id]');
    const regionIdElement = container.closest('[data-region-id]');

    if (!uuidElement || !layoutIdElement || !regionIdElement) {
      return; // Exit if any of the required elements are not found
    }

    const componentUuid = uuidElement.getAttribute('data-uuid');
    const layoutId = regionIdElement
      .closest('[data-lpb-id]')
      .getAttribute('data-lpb-id');
    const regionId = regionIdElement.getAttribute('data-region-id');

    const editableText = container;
    if (!editableText) return; // Exit if the editable text element is not found

    Drupal.ajax({
      url: `${drupalSettings.path.baseUrl}${drupalSettings.path.pathPrefix}mercury-layouts/${layoutId}/edit-title/${componentUuid}/${regionId}`,
      submit: {
        label: editableText.textContent,
      },
    }).execute();
  }

  /**
   * On input update label.
   * @type {(function(...[*]): *)|*}
   */
  const onInput = debounce((event) => {
    saveLabel(event);
  }, 250);

  /**
   * Prevents keypresses from propagating to the component itself.
   *
   * This prevents unintentional component manipulation while typing.
   */
  const containKeypresses = (event) => {
    event.stopPropagation();

    if (event.code === 'Enter' || event.code === 'Escape') {
      event.preventDefault();
      stopEditor();
    }
  };

  Drupal.registerLpbMoveError((settings, element, source, handle) => {
    if (!element) {
      return;
    }
    const component = element.closest('[data-uuid]');
    const layoutType = component.getAttribute('data-layout');
    if (
      (layoutType === 'me_tabs' || layoutType === 'me_accordion') &&
      !handle.classList.contains('lpb-drag')
    ) {
      return 'Tab layouts must be dragged using drag handle.';
    }
  }, 'moves');

  /**
   * Removes a region from the layout.
   *
   * @param {Event} event The event object.
   */
  function removeRegion(event) {
    const componentUuid = event.target
      .closest('[data-uuid]')
      .getAttribute('data-uuid');
    const layoutId = event.target
      .closest('[data-lpb-id]')
      .getAttribute('data-lpb-id');
    const regionId = event.target
      .closest('[data-region-id]')
      .getAttribute('data-region-id');
    Drupal.ajax({
      url: `${drupalSettings.path.baseUrl}${drupalSettings.path.pathPrefix}mercury-layouts/${layoutId}/remove-tab/${componentUuid}/${regionId}`,
      submit: {
        delete: componentUuid,
      },
    }).execute();
  }

  /**
   * Saves the title of a tab.
   *
   * @param {Event} event The event that triggered the save.
   */
  function onDocumentClick(event) {
    if (!event.target.hasAttribute('contenteditable')) {
      document.querySelector('[contenteditable]').blur();
    }
  }

  /**
   * Disables label editing for a region.
   */
  function stopEditor() {
    const container = document.querySelector('[data-editable-label]');
    if (!container) {
      return;
    }
    saveLabel();
    container.closest('.is-editing').classList.remove('is-editing');
    container.removeEventListener('keydown', containKeypresses);
    container.removeEventListener('input', onInput);
    container.removeEventListener('blur', stopEditor);
    if (!container) return; // Exit if the container is not found

    const regionLabel = container.closest('[data-region-id]');
    if (!regionLabel) return; // Exit if the region label is not found

    const btnGroup = regionLabel.querySelector(
      '.me-dynamic-regions-layout__btn-group',
    );
    if (btnGroup) {
      btnGroup.style.display = btnGroup.getAttribute(
        'data-original-style-display',
      );
      btnGroup.removeAttribute('data-original-style-display');
    }

    const btn = regionLabel.querySelector('button[aria-controls]');
    if (!btn) return; // Exit if the button is not found

    const textNode = btn.querySelector('.text') || btn;
    const editableText = container;
    if (editableText) {
      textNode.innerText = editableText.innerText;
    }

    btn.style.display = btn.getAttribute('data-original-style-display');
    btn.removeAttribute('data-original-style-display');

    container.remove();
    document.removeEventListener('mousedown', onDocumentClick);
    document.removeEventListener('keydown', onKeydown);
  }

  /**
   * Handles keydown events while editing.
   *
   * @param {Event} event The event.
   */
  function onKeydown(event) {
    if (event.code === 'Enter' || event.code === 'Escape') {
      event.preventDefault();
      event.stopPropagation();
      stopEditor();
    }
  }

  /**
   * Enables label editing for a region.
   *
   * @param {Element} element  The element to enable the edit mode on.
   */
  function startEditor(element) {
    stopEditor();
    element.classList.add('is-editing');

    // @todo This should be the element that contains the label text, but is
    // it always a button?
    const label = element.querySelector('button');
    const container = document.createElement('div');
    container.className = label.className;
    container.setAttribute('data-editable-label', 'true');
    container.contentEditable = true;
    container.innerText = label.innerText;

    container.addEventListener('keydown', containKeypresses);
    container.addEventListener('input', onInput);
    container.addEventListener('blur', stopEditor);

    label.setAttribute('data-original-style-display', label.style.display);
    label.style.display = 'none';

    label.insertAdjacentElement('beforebegin', container);

    setTimeout(() => {
      container.focus();
      const range = document.createRange();
      const sel = window.getSelection();
      range.selectNodeContents(container);
      sel.removeAllRanges();
      sel.addRange(range);
    }, 0);

    document.addEventListener('mousedown', onDocumentClick);
    document.addEventListener('keydown', onKeydown);
  }

  function editBtnClick(event) {
    startEditor(event.target.closest('[data-region-id]'));
    event.stopPropagation();
    event.preventDefault();
  }

  function deleteBtnClick(event) {
    if (
      // eslint-disable-next-line no-alert
      window.confirm(
        Drupal.t("Delete this tab and it's content? There is no undo."),
      )
    ) {
      removeRegion(event);
    }
    event.stopPropagation();
    event.preventDefault();
  }

  /**
   * Prevents the default action of an event.
   * This is needed to prevent Layout Paragraph's builder.js
   * from moving focus away from the editable label.
   *
   * @param {MouseEvent} event
   * @returns {void}
   */
  function cancelEvent(event) {
    event.stopPropagation();
    event.preventDefault();
  }

  Drupal.behaviors.mercuryEditorTabsEditorUi = {
    attach() {
      once(
        'me-dynamic-regions-layout-btns',
        '.me-dynamic-regions-layout-btn--edit',
      ).forEach((editBtn) => {
        editBtn.addEventListener('mouseup', cancelEvent);
        editBtn.addEventListener('mousedown', editBtnClick);
        editBtn.addEventListener('click', editBtnClick);
      });
      once(
        'me-dynamic-regions-layout-btns',
        '.me-dynamic-regions-layout-btn--delete',
      ).forEach((deleteBtn) => {
        deleteBtn.addEventListener('mouseup', cancelEvent);
        deleteBtn.addEventListener('mousedown', deleteBtnClick);
        deleteBtn.addEventListener('click', deleteBtnClick);
      });
      once(
        'me-dynamic-regions-layout-btns',
        '.me-dynamic-regions-layout-btn--add',
      ).forEach((addBtn) => {
        addBtn.addEventListener('click', (event) => {
          const layoutId = addBtn
            .closest('[data-lpb-id]')
            .getAttribute('data-lpb-id');
          const componentUuid = addBtn
            .closest('[data-uuid]')
            .getAttribute('data-uuid');
          Drupal.ajax({
            url: `${drupalSettings.path.baseUrl}${drupalSettings.path.pathPrefix}mercury-layouts/${layoutId}/add-region/${componentUuid}`,
            submit: {
              add: layoutId,
            },
          }).execute();
          event.stopPropagation();
          event.preventDefault();
        });
      });
      once(
        'me-dynamic-regions-layout-sortable',
        '.me-dynamic-regions-layout-label-group',
      ).forEach((sortableLabels) => {
        Sortable.create(sortableLabels, {
          onChoose(event) {
            const group = event.item.closest('.me-dynamic-regions-layout');
            if (group) {
              group.classList.add('is-dragging');
            }
          },
          onUnchoose(event) {
            const group = event.item.closest('.me-dynamic-regions-layout');
            if (group) {
              group.classList.remove('is-dragging');
            }
          },
          // Changed sorting within list.
          onUpdate(event) {
            const element = event.item;
            const container = element.closest(
              '.me-dynamic-regions-layout-label-group',
            );
            setTimeout(() => {
              const layoutId = container
                .closest('[data-lpb-id]')
                .getAttribute('data-lpb-id');
              const componentUuid = container
                .closest('[data-uuid]')
                .getAttribute('data-uuid');
              const labels = [
                ...container.querySelectorAll('[data-region-id]'),
              ];
              const order = labels.map((label) =>
                label.getAttribute('data-region-id'),
              );
              Drupal.ajax({
                url: `${drupalSettings.path.baseUrl}${drupalSettings.path.pathPrefix}mercury-layouts/${layoutId}/reorder/${componentUuid}`,
                submit: {
                  order: JSON.stringify(order),
                },
              }).execute();
            }, 200);
          },
        });
      });
    },
  };
})(Drupal, Drupal.debounce, Sortable, once);
